#!/usr/bin/python
import os
import sys
import config
import up2dateLog
import up2dateErrors 
import up2dateAuth
import up2dateUtils 
import wrapperUtils
import socket
import string
import time


from rhn import rpclib
    
from translate import _, N_, cat
            
# get a reference to the log file object
log = up2dateLog.initLog()
            

#grab an instance of the config class shared
cfg = config.initUp2dateConfig()

warningCallback = wrapperUtils.warningCallback

# Where do we keep the CA certificate for RHNS?
# The servers we're talking to need to have their certs
# signed by one of these CA.
if cfg.readEntry("sslCACert"):
    rhns_ca_cert = cfg.readEntry("sslCACert")
else:
    rhns_ca_cert = "/usr/share/rhn/RHNS-CA-CERT"

def hasSSL():
    return hasattr(socket, "ssl")

class RetryServer(rpclib.Server):
    def foobar(self):
        pass
                                                                                
    def addServerList(self, serverList):
        self.serverList = serverList
                                                                                
    def _request1(self, methodname, params):
        while 1:
            try:
                ret = self._request(methodname, params)
            except rpclib.InvalidRedirectionError:
                raise
                                                                                
            except:
                server = self.serverList.next()
                if server == None:
                    print "We are out of serverUrl's to try. We tried:"
                    for i in self.serverList.serverList:
                        print i
                    raise
                                                                                
                msg = "An error occured talking to %s:\n" % self._host
                msg = msg + "%s\n%s\n" % (sys.exc_type, sys.exc_value)
                msg = msg + "Trying the next serverURL: %s\n" % self.serverList.server()
                log.log_me(msg)
                # try a different url
                                                                                
                # use the next serverURL
                import urllib
                typ, uri = urllib.splittype(self.serverList.server())
                typ = string.lower(typ)
                if typ not in ("http", "https"):
                    raise InvalidRedirectionError(
                        "Redirected to unsupported protocol %s" % typ)
                                                                                
                self._host, self._handler = urllib.splithost(uri)
                self._orig_handler = self._handler
                self._type = typ
                if not self._handler:
                    self._handler = "/RPC2"
                self._allow_redirect = 1
                continue
            # if we get this far, we succedded
            break
        return ret
                                                                                
                                                                                
    def __getattr__(self, name):
        # magic method dispatcher
        return rpclib.transports.xmlrpclib._Method(self._request1, name)
                                                                                


# uh, yeah, this could be an iterator, but we need it to work on 1.5 as well
class ServerList:
    def __init__(self, serverlist=[]):
        self.serverList = serverlist
        self.index = 0
                                                                                
    def server(self):
        self.serverurl = self.serverList[self.index]
        return self.serverurl
                                                                                
    def next(self):
        self.index = self.index + 1
        if self.index >= len(self.serverList):
            return None
        return self.server()
                                                                                
    def resetServerList(self, serverlist):
        self.serverList = serverlist
        self.index = 0
                                                                                

# singleton for the ServerList
def initServerList(servers):
    global server_list
    try:
        server_list = server_list
    except NameError:
        server_list = None
                                                                                
    if server_list == None:
        server_list = ServerList(servers)

    # if we've changed the config, we need need to
    # update the server_list as well. Not really needed
    # in the app, but makes testing cleaner
    sl = cfg.readEntry("serverURL")
    if type(sl) == type(""):
        sl  = [sl]
                                                                                
    if sl != server_list.serverList:
        server_list = ServerList(servers)
                                                                                
    return server_list




def getServer(refreshCallback=None):
    global warningCallback


# Where do we keep the CA certificate for RHNS?
# The servers we're talking to need to have their certs
# signed by one of these CA.
    ca = cfg.readEntry("sslCACert")
    if type(ca) == type(""):
        ca = [ca]
                                                                                                                       
    rhns_ca_certs = ca or ["/usr/share/rhn/RHNS-CA-CERT"]

        
    if cfg.readEntry("enableProxy"):
        proxyHost = up2dateUtils.getProxySetting()
    else:
        proxyHost = None

    if hasSSL():
        serverUrls = cfg.readEntry("serverURL")
    else:
        serverUrls = cfg.readEntry("noSSLServerURL")

    # the standard is to be a string, so list-a-fy in that case
    if type(serverUrls) == type(""):
        serverUrls = [serverUrls]


    proxyUser = None
    proxyPassword = None
    if cfg.readEntry("enableProxyAuth"):
        proxyUser = cfg.readEntry("proxyUser") or None
        proxyPassword = cfg.readEntry("proxyPassword") or None

    lang = None
    for env in 'LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG':
        if os.environ.has_key(env):
            if not os.environ[env]:
                # sometimes unset
                continue
            lang = string.split(os.environ[env], ':')[0]
            lang = string.split(lang, '.')[0]
            break

    serverList = initServerList(serverUrls)
    
    s = RetryServer(serverList.server(),
                    refreshCallback=refreshCallback,
                    proxy=proxyHost,
                    username=proxyUser,
                    password=proxyPassword)
    s.addServerList(serverList)
    s.add_header("X-Up2date-Version", up2dateUtils.version())
#    s = rpclib.Server(serverUrl, proxyHost,
#			 transport=None,
#                         refreshCallback=refreshCallback,
#                         username=proxyUser,
#                         password=proxyPassword)
    if lang:
        s.setlang(lang)

    # force the validation of the SSL cert
   # require RHNS-CA-CERT file to be able to authenticate the SSL connections
    for rhns_ca_cert in rhns_ca_certs:
        if not os.access(rhns_ca_cert, os.R_OK):
            log.log_me("ERROR: can not find RHNS CA file: %s" % rhns_ca_cert)
            sys.exit(-1)
                                                                                                                       
        # force the validation of the SSL cert
        s.add_trusted_cert(rhns_ca_cert)


    return s


# FIXME: doCall should probabaly be a method
# of a higher level server object
def doCall(method, *args, **kwargs):
    ret = None

    attempt_count = 1
    if cfg.readEntry("networkRetries"):
        attempts = cfg.readEntry("networkRetries")
    else:
        attempts = 5
    
    while 1:
        failure = 0
        ret = None
        try:
            ret = apply(method, args, kwargs)
        except KeyboardInterrupt:
            raise up2dateErrors.CommunicationError(_("Connection aborted by the user"))
        # if we get a socket error, keep trying
        except (socket.error, socket.sslerror), e:
            log.log_me("A socket error occurred: %s, attempt #%s" % (e, attempt_count))
            if attempt_count >= attempts:
                if len(e.args) > 1:
                    raise up2dateErrors.CommunicationError(e.args[1])
                else:
                    raise up2dateErrors.CommunicationError(e.args[0])
            else:
                failure = 1
        except rpclib.ProtocolError, e:
            log.log_me("A protocol error occurred: %s , attempt #%s" % (e.errmsg, attempt_count))
            (errCode, errMsg ) = rpclib.reportError(e.headers)
            reset = 0
            if abs(errCode) == 34:
                log.log_me("Auth token timeout occurred\n errmsg: %s" % errMsg)
                # this calls login, which in tern calls doCall (ie,
                # this function) but login should never get a 34, so
                # should be safe from recursion

                # FIXME: move to up2dateAuth
                up2dateAuth.updateLoginInfo()
                reset = 1

            # the servers are being throttle to pay users only, catch the
            # exceptions and display a nice error message
            if abs(errCode) == 51:
                log.log_me(_("Server was refused connection due to high load"))
                raise up2dateErrors.CommunicationError(e.errmsg)


            # if we get a 404 from our server, thats pretty fatal... no point in
            # retrying over and over. Note that errCode == 17 is specific to
            # our servers, if the serverURL is just pointing somewhere random
            # they will get a 0 for errcode and will raise a CommunicationError
            if abs(errCode) == 17:
                if type(args[0]) == type([]):
                    pkg = args[0]
                else:
                    pkg=args[1]
                                                                                                                                  
                if type(pkg) == type([]):
                    pkgName = "%s-%s-%s.%s" % (pkg[0], pkg[1], pkg[2], pkg[4])
                else:
                    pkgName = pkg
                msg = "File Not Found: %s\n%s" % (pkgName, errMsg)
                log.log_me(msg)

                raise up2dateErrors.FileNotFoundError(msg)
                
            if not reset:
                if attempt_count >= attempts:
                    raise up2dateErrors.CommunicationError(e.errmsg)
                else:
                    failure = 1
            
        except rpclib.ResponseError:
            raise up2dateErrors.CommunicationError("Broken response from the server.");

	if attempt_count >= attempts:
	    raise up2dateErrors.CommunicationError("The data returned from the server was incomplete")

	if ret != None:
	    break
	# were not supposed to get None here, so thats a failure
	else:
	    failure = 1

        if failure:
            # rest for five seconds before trying again
            time.sleep(5)
            attempt_count = attempt_count + 1
        
    return ret
    
